--[[
Copyright 2010, 2011 William Matthews. All rights reserved.

See LICENSE for license terms.
]]

local CB = select(2,...)
local P = CB.Private
local Events = P:NewModule("Events", "AceEvent-3.0")

local Queue = P.Queue
local setdefault = P.setdefault
local Timer = P.Timer

local CallbackHandler = {}
local CallbackHandlerMT = {__index=CallbackHandler}
local function timeoutCB(cb)
	cb.__timeout=nil
	cb:Fail()
end
function CallbackHandler.New(success,fail,timeout,cleanup)
	local rv={}
	rv.__success=success
	rv.__fail=fail
	rv.__valid=true
	rv.__timeout=timeout and Timer.new(timeoutCB,timeout,rv)
	rv.__cleanup=cleanup
	setmetatable(rv,CallbackHandlerMT)
	return rv
end
function CallbackHandler:Success(...)
	if self.__success then
		self.__success(...)
	end
	self:Clear()
end
function CallbackHandler:Fail(...)
	if self.__fail then
		self.__fail(...)
	end
	self:Clear()
end
function CallbackHandler:Valid()
	return self.__valid
end
function CallbackHandler:Clear()
	self.__success=nil
	self.__fail=nil
	self.__valid=false
	if self.__timeout then
		self.__timeout:Cancel()
		self.__timeout=nil
	end
	if self.__cleanup then
		self.__cleanup()
		self.__cleanup=nil
	end
end
function CallbackHandler:UpdateTimeout(time)
	if self.__valid then
		self.__timeout = Timer.updateOrNew(self.__timeout,timeoutCB,time,self)
	end
end

local spellHandlers={} -- map spell name to table of handlers
local itemHandlers={} -- map item ID to table of handlers

--local spells = {}
local spellCBs = {}
setdefault(spellCBs,Queue.New)

local itemLink = nil
local itemBound = nil
local itemTimestamp = nil
local itemCB = nil

local lootWindow = nil
local lootCB = nil

local createItem = nil
local createCount = nil
local createTimestamp = nil
local createTimer = nil
local createCB = nil

local receiveCbs_ = {}
setdefault(receiveCbs_, Queue.New)
local receiveMultipleCbs_ = Queue.New() 

local function callback(f,...)
	if f then
		return f(...)
	end
end

local function itemCBCleanup()
	itemLink = nil
	itemBound = nil
	itemTimestamp = nil
	itemCB = nil
end	

local function doItem(link,bound)
	if not link then return end
	if itemCB then
		itemCB:Success(link,bound)
	else
		local id = P.idOfItem(link)
		if itemHandlers[id] then
			for _, h in ipairs(itemHandlers[id]) do
				h(link,bound)
			end
		else
			itemLink = link
			itemBound = bound
			itemTimestamp = GetTime()
		end
	end
end

local function doCreateCBSuccess()
	if createCB then
		createCB:Success(createItem,createCount)
	end
	createItem = nil
	createCount = nil
	createTimestamp = nil
	createTimer = nil
	createCB = nil
end

local function doCreateTimer()
	createTimer = Timer.updateOrNew(createTimer,doCreateCBSuccess,2*P.EPSILON_TIME)
end

--[[
local function doReceive(id)
	while receiveItems[id] and receiveItems[id].count>0 and not receiveCBs[id]:Empty() do
		local cb=receiveCBs[id]:Pop()
		if cb:Valid() then
			if receiveItems[id].count>=cb.count then
				receiveItems[id].count=receiveItems[id].count-cb.count
				cb:Success(cb.item,cb.count)
			else
				cb:Fail()
			end
		end
	end
	receiveTimers[id] = nil
end

local function doReceiveTimer(id)
	receiveTimers[id]=Timer.updateOrNew(receiveTimers[id],doReceive,2*P.EPSILON_TIME,id)
end
--]]

---------
-- API --
---------

-- permanent handlers

-- calls handler(spellName) when spellName STARTS casting, use SpellCallback to check for success
function P:AddSpellHandler(spellName,handler)
	if not spellHandlers[spellName] then
		spellHandlers[spellName] = {}
	end
	table.insert(spellHandlers[spellName],handler)
end

-- calls handler(item) when an item is used
function P:AddItemHandler(item,handler)
	local id = self.idOfItem(item)
	if not itemHandlers[id] then
		itemHandlers[id] = {}
	end
	table.insert(itemHandlers[id],handler)
end

-- one-shot handlers

-- calls successCB() if the next or current spell cast is spellName, otherwise calls failCB
function P:SpellCallback(spellName,successCB,failCB)
	local cb = CallbackHandler.New(successCB,failCB)
	spellCBs[spellName]:Push(cb)
	local name, subText, _, _, startTime, endTime, isTradeSkill, castID, notInterruptible = UnitCastingInfo("player")
	if name and name==spellName then
		cb:UpdateTimeout(endTime / 1000 + P.TIMEOUT_TIME)
	else
		local _, _, _, _, _, _, castingTime, _, _ = GetSpellInfo(spellName)
		if not castingTime then
			Timer.new(function ()
				if not cb:Valid() then return end
				local name, subText, _, _, startTime, endTime, isTradeSkill, castID, notInterruptible = UnitCastingInfo("player")
				if name and name==spellName then
					cb:UpdateTimeout(endTime / 1000 + P.TIMEOUT_TIME)
				else
					cb:Fail()
				end
			end,P.TIMEOUT_TIME)
		else
			cb:UpdateTimeout(castingTime / 1000 + P.TIMEOUT_TIME)
		end
	end
end

-- calls successCB(item) when an item is used, failCB if an item doesn't get used
function P:ItemCallback(successCB,failCB)
	if itemCB and itemCB:Valid() then
		self:DebugPrint("Multiple item callbacks at once.  Failing oldest.")
		itemCB:Fail()
	end
	if itemLink and math.abs(GetTime() - itemTimestamp)<=P.EPSILON_TIME then
		if successCB then
			successCB(itemLink,itemBound)
		end
		itemCBCleanup()
	else
		itemCB = CallbackHandler.New(successCB,failCB,nil,itemCBCleanup) -- we don't necessarily know how long to wait for an item... check for spell targetting based events?
	end
end

--[[ -- NO LONGER WORKS

-- calls successCB(item,count) when an item is created
function P:CreateCallback(successCB,failCB)
	if createCB and createCB:Valid() then
		createCB:Fail()
	end
	if createItem and math.abs(GetTime() - createTimestamp)<=P.EPSILON_TIME then
		doCreateTimer()
	end
	createCB = CallbackHandler.New(successCB,failCB,P.TIMEOUT_TIME)
end
]]

function P:ReceiveCallback(id,count,successCB,failCB)
	assert(type(id)=="number")
	local cb=CallbackHandler.New(successCB, failCB, P.TIMEOUT_TIME)
	cb.item=item
	cb.count=count
	receiveCbs_[id]:Push(cb)
end

-- calls successCB(lootTbl) where lootTbl is {id={item,count,[lost]}, ...}
function P:ReceiveMultipleCallback(successCb, failCb)
  receiveMultipleCbs_:Push(CallbackHandler.New(successCb, failCb, P.TIMEOUT_TIME))
end

-- calls successCB(lootTbl) where lootTbl is {id={item,count,[lost]}, ...}
function P:LootCallback(successCB,failCB)
	if lootCB and lootCB:Valid() then
		lootCB:Fail()
	end
	lootCB = CallbackHandler.New(successCB,failCB) -- ,P.TIMEOUT_TIME) -- don't timeout loot!  TODO: timeout if loot doesn't _open_, but then give time to actually loot
end

--------------------
-- IMPLEMENTATION --
--------------------

-- tooltip for testing soulbound-ness
local SBTT = CreateFrame("GameTooltip","CB_SBTT",UIParent,"GameTooltipTemplate")
local SBTT2 = _G["CB_SBTTTextLeft2"]
local SBTT3 = _G["CB_SBTTTextLeft3"]
local BOUND = {
	[ITEM_SOULBOUND] = ITEM_SOULBOUND,
	[ITEM_BIND_QUEST] = ITEM_BIND_QUEST,
	[ITEM_BIND_ON_PICKUP] = ITEM_BIND_ON_PICKUP,
	[ITEM_CONJURED] = ITEM_CONJURED,
	[ITEM_ACCOUNTBOUND] = ITEM_ACCOUNTBOUND,
	[ITEM_BIND_TO_ACCOUNT] = ITEM_BIND_TO_ACCOUNT,
	[ITEM_BIND_TO_BNETACCOUNT] = ITEM_BIND_TO_BNETACCOUNT,
	[ITEM_BNETACCOUNTBOUND] = ITEM_BNETACCOUNTBOUND,
}

local function getBound(ttMethod,...)
	SBTT:SetOwner(UIParent,"ANCHOR_NONE")
	SBTT:ClearLines()
	SBTT[ttMethod](SBTT,...)
	local rv = BOUND[SBTT2:GetText()] or BOUND[SBTT3:GetText()]
	SBTT:Hide()
	return rv
end

local function Hook_UseContainerItem(bag,slot)
	doItem(GetContainerItemLink(bag,slot),getBound("SetBagItem",bag,slot))
end

local function Hook_PickupInventoryItem(slot)
	doItem(GetInventoryItemLink("player",slot),getBound("SetInventoryItem","player",slot))
end

local function Hook_SpellTargetItem(itemString)
	local _, link = GetItemInfo(itemString)
	local id = P.idOfItem(link)
	doItem(link,getBound("SetItemByID",id))
end

local function Hook_UseItemByName(itemString)
	local _, link = GetItemInfo(itemString)
	local id = P.idOfItem(link)
	doItem(link,getBound("SetItemByID",id))
end

local function Event_UNIT_SPELLCAST_SUCCEEDED(event,unit,spell,rank)
	if unit~="player" then return end
	--CB:Print("SUCCESS: "..spell)
	local cb = spellCBs[spell] and spellCBs[spell]:Pop()
	if cb then
		cb:Success(spell)
	end
end

local function Event_UNIT_SPELLCAST_SENT(event,unit,spell,rank,target)
	if unit~="player" then return end
	--CB:Print("SENT: "..spell)
--	spells[spell]=GetTime()
	local hs = spellHandlers[spell] or {}
	for _, h in ipairs(hs) do
		h(spell)
	end
end

local function Event_UNIT_SPELLCAST_FAILED(event,unit,spell,rank)
	if unit~="player" then return end
	local cb=spellCBs[spell]:Pop()
	if cb then
		cb:Fail()
	end
end

local function Event_UNIT_SPELLCAST_INTERRUPTED(event,unit,spell,rank)
	if unit~="player" then return end
	local cb=spellCBs[spell]:Pop()
	while cb and not cb:Valid() do
		cb=spellCBs[spell]:Pop()
	end
	if cb then
		cb:Fail()
	end
end

local function Event_UI_ERROR_MESSAGE(event,msg)
	if msg=="Inventory is full." then
	end
end

--[[
local function DoReceiveItem(id, count)
  --P:DebugPrint({"DoReceiveItem", id, count})
  if not receiveItems[id] or math.abs(GetTime() - receiveItems[id].timestamp)>P.EPSILON_TIME then
  	receiveItems[id]={count=0}
  end
  receiveItems[id].count=receiveItems[id].count + count
  receiveItems[id].timestamp=GetTime()
  if receiveCBs[id] and not receiveCBs[id]:Empty() then
  	doReceiveTimer(id)
  end
end
--]]

local function Event_CHAT_MSG_LOOT(event,msg,sender,lang,channelString,target,flags,unknown,channelNumber)
	-- Create items
	local m,c = msg:match("^You create: (.+)x(%d+)%.$")
	if not m then
		m=msg:match("^You create: (.+)%.$")
		c=1
	else
		c=tonumber(c)
	end
	if m then
		local id=CB.Private.idOfItem(m)
		if createTimestamp and math.abs(GetTime() - createTimestamp)<=P.EPSILON_TIME then
			if createItem~=id then
				P:DebugPrint("Created two different items within "..P.EPSILON_TIME.." seconds! BUG?")
			else
				createCount = createCount + c				
			end
		else
			createItem = id
			createCount = c
		end
		createTimestamp = GetTime()
		doCreateTimer()
		return
	end
	
	-- Receive items
	m,c=msg:match("^You receive item: (.+)x(%d+)%.$")
	if not m then
		m=msg:match("^You receive item: (.+)%.$")
		c=1
	else
		c=tonumber(c)
	end
	if m then
		local id=CB.Private.idOfItem(m)
--		DoReceiveItem(id, c)
	end
	
end

local function Event_LOOT_OPENED(event,autoloot)
	local numLoot = GetNumLootItems()
	lootWindow = {}
	for i = 1,numLoot do
		if GetLootSlotType(i) == 1 then -- item
			local _, _, quantity = GetLootSlotInfo(i)
			local id = P.idOfItem(GetLootSlotLink(i))
			lootWindow[i] = {
				item = id,
				count = quantity,
			}
		elseif GetLootSlotType(i) == 2 then -- money
			local _, item = GetLootSlotInfo(i)
			local money = tonumber(item:match("(%d+) Gold") or "0")*10000 + tonumber(item:match("(%d+) Silver") or "0")*100 + tonumber(item:match("(%d+) Copper") or "0")
--			P:Print(format("Loot window: %s",P.moneyString(money)))
--			CB:DebugPrint(item)
		elseif GetLootSlotType(i) == 3 then -- currency
			-- ignore (for now?  forever?)
		else
			P:DebugPrint(format("Loot slot %d is neither coin nor item nor currency",i))
		end
	end
end

local function Event_LOOT_SLOT_CLEARED(event,slot) -- TODO: right now this assumes that the slot was cleared by the player rather than a raid/party memeber... that's ok for DE/prospect/milling
--  P:DebugPrint({event=event, slot=slot})
	if lootWindow and lootWindow[slot] then
		lootWindow[slot].looted = true
	else
		-- this happens in the case of gold & currency!
--		CB:DebugPrint({event=event,lootWindow=lootWindow})
	end
end

local function Event_LOOT_SLOT_CHANGED(event, slot, ...)
--  P:DebugPrint({event=event, slot=slot, args={...}})
end

local function DoLootClosed_()
--  P:DebugPrint(lootWindow)
  if lootWindow then
  	local items = P.ItemSet.new()
  	local lost = P.ItemSet.new()

  	for _, item in pairs(lootWindow) do
  		local id = P.idOfItem(item.item)
  		if item.looted then
  			items:Inc(id,item.count)
  		else
  			lost:Inc(id,item.count)
  		end
  	end

  	if lootCB then
  		lootCB:Success(items,lost)
  	end
  	lootWindow = nil
  else
  	--CB:DebugPrint({event=event,lootWindow=lootWindow})
  end	
end

local function Event_LOOT_CLOSED(event)
  Timer.new(DoLootClosed_, 20 * P.EPSILON_TIME)
end

local bag_cache_id_ = {}
local bag_cache_count_ = {}
local bag_update_timer_ = nil

local function IncrementBagUpdates_(id, delta)
  bag_updates[id] = (bag_updates[id] or 0) + delta
  if bag_updates[id] == 0 then
    bag_updates[id] = nil
  end
end

local function DoBagUpdates_()
  local bag_updates = {}
  
  ---- Scan Bags ----
  for bag_id = 0, 4 do
    local size = GetContainerNumSlots(bag_id)
    if not bag_cache_id_[bag_id] then
      bag_cache_id_[bag_id] = {}
      bag_cache_count_[bag_id] = {}
    end 
    for i = 1, size do
      local _, count, _, _, _, _, item = GetContainerItemInfo(bag_id, i)
      local id = item and P.idOfItem(item)
      count = count or 0
      if bag_cache_count_[bag_id][i] then
        -- We have valid previous data.
        local prev_id = bag_cache_id_[bag_id][i]
        local prev_count = bag_cache_count_[bag_id][i]
        --P:DebugPrint({bag_id = bag_id, i = i, id = id, count = count, prev_id = prev_id, prev_count = prev_count})
        
        if id then bag_updates[id] = (bag_updates[id] or 0) + count end
        if prev_id then bag_updates[prev_id] = (bag_updates[prev_id] or 0) - prev_count end
      end
      bag_cache_count_[bag_id][i] = count
      bag_cache_id_[bag_id][i] = id
    end
  end
  
  ---- Process Events ----
  local free = {}
  
  for _, item in pairs(lootWindow or {}) do
    if not item.looted and bag_updates[item.item] and bag_updates[item.item] >= item.count then
      item.looted = true
      bag_updates[item.item] = bag_updates[item.item] - item.count
--      P:DebugPrint("looted")
    end
  end
  
  for id, count in pairs(bag_updates) do  
    local q = receiveCbs_[id]
    if q:Empty() then
      if count > 0 then free[id] = count end
    else
      while not q:Empty() and (not q:Head():Valid() or q:Head().count <= count) do
        local cb = q:Pop()
        if cb:Valid() then
          count = count - cb.count
          cb:Success()
        end
      end
      if count > 0 then
        free[id] = count
      end
    end
  end
  
  while not receiveMultipleCbs_:Empty() do
    local cb = receiveMultipleCbs_:Pop()
    if cb:Valid() then
--      P:DebugPrint({bag_updates, free})
      cb:Success(free)
      break
    end
  end
end

local function Event_BAG_UPDATE(event, bag_id)
  bag_update_timer_ = Timer.updateOrNew(bag_update_timer_, DoBagUpdates_, 2 * P.EPSILON_TIME)
end

function Events:OnInitialize()
	hooksecurefunc("UseContainerItem", Hook_UseContainerItem)
	hooksecurefunc("PickupInventoryItem", Hook_PickupInventoryItem)
	hooksecurefunc("SpellTargetItem", Hook_SpellTargetItem)
	hooksecurefunc("UseItemByName", Hook_UseItemByName)
	
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED",Event_UNIT_SPELLCAST_SUCCEEDED)
	self:RegisterEvent("UNIT_SPELLCAST_SENT",Event_UNIT_SPELLCAST_SENT)
	self:RegisterEvent("UNIT_SPELLCAST_FAILED",Event_UNIT_SPELLCAST_FAILED)
	self:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED",Event_UNIT_SPELLCAST_INTERRUPTED)
	self:RegisterEvent("UI_ERROR_MESSAGE",Event_UI_ERROR_MESSAGE)
	self:RegisterEvent("CHAT_MSG_LOOT",Event_CHAT_MSG_LOOT)
	self:RegisterEvent("LOOT_OPENED",Event_LOOT_OPENED)
	self:RegisterEvent("LOOT_SLOT_CLEARED",Event_LOOT_SLOT_CLEARED)
	self:RegisterEvent("LOOT_SLOT_CHANGED",Event_LOOT_SLOT_CHANGED)
	self:RegisterEvent("LOOT_CLOSED",Event_LOOT_CLOSED)
	self:RegisterEvent("BAG_UPDATE", Event_BAG_UPDATE)
	
  bag_update_timer_ = Timer.updateOrNew(bag_update_timer_, DoBagUpdates_, 2 * P.EPSILON_TIME)	
end
